from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, List, Literal, Optional

from contaxy.schema import Job, JobInput, ResourceAction, Service, ServiceInput
from contaxy.schema.deployment import DeploymentType, ServiceUpdate
from contaxy.schema.shared import ResourceActionExecution


class ServiceOperations(ABC):
    @abstractmethod
    def list_services(
        self,
        project_id: str,
        deployment_type: Literal[
            DeploymentType.SERVICE, DeploymentType.EXTENSION
        ] = DeploymentType.SERVICE,
    ) -> List[Service]:
        """Lists all services associated with the given project.

        Args:
            project_id (str): The project ID to filter the services.
            deployment_type (One of [DeploymentType.SERVICE, DeploymentType.JOB]): The deployment type of either Service or Extension (which is a subtype of Service).

        Returns:
            List[Service]: The list of services associated with the project.
        """
        pass

    @abstractmethod
    def deploy_service(
        self,
        project_id: str,
        service_input: ServiceInput,
        action_id: Optional[str] = None,
        deployment_type: Literal[
            DeploymentType.SERVICE, DeploymentType.EXTENSION
        ] = DeploymentType.SERVICE,
        wait: bool = False,
    ) -> Service:
        """Deploys a service for the specified project.

        If no `action_id` is provided, the system will automatically select the best deployment option.

        Available deployment options (actions) can be requested via the [list_deploy_service_actions](#services/list_deploy_service_actions) operation.
        If the action is from an extension, the `action_id` must be a composite ID with the following format: `{extension_id}~{action_id}`.

        The action mechanism is further explained in the description of the [list_deploy_service_actions](#services/list_deploy_service_actions).

        Args:
            project_id (str): The id of the project that the service should be assigned to.
            service_input (ServiceInput): The service input which can be used to configure the deployed service.
            action_id (Optional[str], optional): The ID of the selected action. Defaults to `None`.
            deployment_type (One of [DeploymentType.SERVICE, DeploymentType.JOB]): The deployment type of either Service or Extension (which is a subtype of Service).
            wait (bool, optional): If set to True, the function will wait until the service was successfully created.

        Returns:
            Service: The metadata of the deployed service.
        """
        pass

    @abstractmethod
    def list_deploy_service_actions(
        self,
        project_id: str,
        service: ServiceInput,
    ) -> List[ResourceAction]:
        """Lists all available service deployment options (actions).

        Args:
            project_id (str): The project ID associated with the service.
            service_id (str): The ID of the service.

        Returns:
            List[ResourceAction]: Available deployment actions.
        """
        pass

    @abstractmethod
    def get_service_metadata(
        self,
        project_id: str,
        service_id: str,
    ) -> Service:
        """Returns the metadata of a single service.

        Args:
            project_id (str): The project ID associated with the service.
            service_id (str): The ID of the service.

        Returns:
            Service: The service metadata.
        """
        pass

    @abstractmethod
    def update_service(
        self, project_id: str, service_id: str, service: ServiceUpdate
    ) -> Service:
        """Updates the service.

        Args:
            project_id (str): The project ID associated with the service.
            service_id (str): The ID of the service.
            service (ServiceUpdate): Updates that should be applied to the service
        Returns:
            Service: The updated service metadata
        """
        pass

    @abstractmethod
    def update_service_access(self, project_id: str, service_id: str) -> None:
        """Updates the last time the service was accessed and by which user.

        Args:
            project_id (str): The project ID associated with the service.
            service_id (str): The ID of the service.
        """
        pass

    @abstractmethod
    def delete_service(
        self,
        project_id: str,
        service_id: str,
        delete_volumes: bool = False,
    ) -> None:
        """Deletes a service.

        Args:
            project_id (str): The project ID associated with the service.
            service_id (str): The ID of the service.
            delete_volumes (bool, optional): If `True`, all attached volumes will be deleted. Defaults to `False`.

        Raises:
            RuntimeError: If an error occurs during the deletion of the service.
        """
        pass

    @abstractmethod
    def delete_services(
        self,
        project_id: str,
    ) -> None:
        """Deletes all services associated with a project.

        Args:
            project_id (str): The project ID.
        """
        pass

    @abstractmethod
    def get_service_logs(
        self,
        project_id: str,
        service_id: str,
        lines: Optional[int],
        since: Optional[datetime],
    ) -> str:
        """Returns the logs of a service.

        Args:
            project_id (str): The ID of the project into which the service is deployed in.
            service_id (str): The ID of the service.
            lines (Optional[int]): If provided, just the last `n` lines are returned from the log. Defaults to `None`.
            since (Optional[datetime]): If provided, just the logs since the given timestamp are returned. Defaults to `None`.

        Raises:
            NotImplementedError: [description]
            RuntimeError: If reading the logs of the given service fails.

        Returns:
            str: The logs of the service.
        """
        pass

    @abstractmethod
    def suggest_service_config(
        self,
        project_id: str,
        container_image: str,
    ) -> ServiceInput:
        """Suggests an input configuration based on the provided `container_image`.

        The suggestion is based on metadata extracted from the container image (e.g. labels)
        as well as suggestions based on previous project deployments with the same image.

        Args:
            project_id (str): The project ID associated with the service.
            container_image (str): The container image to use as context for the suggestion.

        Returns:
            ServiceInput: The suggested service configuration.
        """
        pass

    @abstractmethod
    def list_service_actions(
        self,
        project_id: str,
        service_id: str,
    ) -> List[ResourceAction]:
        """Lists all actions available for the specified service.

        See the endpoint documentation for more information on the action mechanism.

        Args:
            project_id (str): The project ID associated with the service.
            service_id (str): The ID of the service.

        Returns:
            List[ResourceAction]: Available actions for given services.
        """
        pass

    @abstractmethod
    def execute_service_action(
        self,
        project_id: str,
        service_id: str,
        action_id: str,
        action_execution: ResourceActionExecution = ResourceActionExecution(),
    ) -> Any:
        """Executes the selected service action.

        The actions need to be first requested from the list_service_actions operation.
        If the action is from an extension, the `action_id` must be a composite ID with the following format: `{extension_id}~{action_id}`.

        Args:
            project_id (str): The project ID associated with the service.
            service_id (str): The ID of the service.
            action_id (str): The ID of the selected action.
            action_execution (ResourceActionExecution): The action execution request which contains the action parameters

        Returns:
            `None` or a redirect response to another URL.
        """
        pass


class JobOperations(ABC):
    @abstractmethod
    def list_jobs(self, project_id: str) -> List[Job]:
        pass

    @abstractmethod
    def deploy_job(
        self,
        project_id: str,
        job_input: JobInput,
        action_id: Optional[str] = None,
        wait: bool = False,
    ) -> Job:
        pass

    @abstractmethod
    def list_deploy_job_actions(
        self,
        project_id: str,
        job: JobInput,
    ) -> List[ResourceAction]:
        pass

    @abstractmethod
    def suggest_job_config(self, project_id: str, container_image: str) -> JobInput:
        pass

    @abstractmethod
    def get_job_metadata(self, project_id: str, job_id: str) -> Job:
        pass

    @abstractmethod
    def delete_job(self, project_id: str, job_id: str) -> None:
        pass

    @abstractmethod
    def delete_jobs(
        self,
        project_id: str,
    ) -> None:
        """Deletes all jobs associated with a project.

        Args:
            project_id (str): The project ID.
        """
        pass

    @abstractmethod
    def get_job_logs(
        self,
        project_id: str,
        job_id: str,
        lines: Optional[int] = None,
        since: Optional[datetime] = None,
    ) -> str:
        pass

    @abstractmethod
    def list_job_actions(
        self,
        project_id: str,
        job_id: str,
    ) -> List[ResourceAction]:
        pass

    @abstractmethod
    def execute_job_action(
        self,
        project_id: str,
        job_id: str,
        action_id: str,
        action_execution: ResourceActionExecution = ResourceActionExecution(),
    ) -> Any:
        pass


class DeploymentOperations(ServiceOperations, JobOperations, ABC):
    pass



import os
from datetime import timedelta
from enum import Enum
from typing import List, Optional, Union

from loguru import logger
from pydantic import AnyHttpUrl, BaseSettings, PostgresDsn, validator

API_TOKEN_NAME: str = "ct_token"
AUTHORIZED_USER_COOKIE: str = "ctxy_authorized_user"
SYSTEM_INTERNAL_PROJECT: str = "system"
SYSTEM_ADMIN_USERNAME: str = "admin"
SYSTEM_ADMIN_INITIAL_PASSWORD: str = "admin"


class DeploymentManager(str, Enum):
    KUBERNETES = "kubernetes"
    DOCKER = "docker"


# MAX_SYSTEM_NAMESPACE_LENGTH = 5
# def get_system_namespace(system_name: str) -> str:
#    return id_utils.generate_readable_id(
#        system_name,
#        max_length=MAX_SYSTEM_NAMESPACE_LENGTH,
#        min_length=3,
#        max_hash_suffix_length=MAX_SYSTEM_NAMESPACE_LENGTH,
#        separator="",
#    )


class Settings(BaseSettings):
    """Platform Settings."""

    # TODO Decide on default values
    CONTAXY_HOST: Optional[str] = None
    # Set the base url prefix
    CONTAXY_BASE_URL: str = ""
    # ? Maybe mobe up to system constants for now
    # TODO Make actually configurable, i.e. ensure no hardcoded paths exists anymore (backend/webapp/nginx)
    CONTAXY_API_PATH: str = "api"
    CONTAXY_WEBAPP_PATH: str = "app"

    # TODO: handle HTTP/HTTPS
    # bypass Contaxy nginx and access FastAPI directly
    CONTAXY_API_ENDPOINT: str = "http://contaxy:8090/api"

    # The system namespace used to managed different versions
    SYSTEM_NAMESPACE: str = "ctxy"

    # Selected deployment manager
    DEPLOYMENT_MANAGER: DeploymentManager = DeploymentManager.DOCKER
    KUBERNETES_NAMESPACE: Optional[str] = None
    HOST_DATA_ROOT_PATH: Optional[str] = None
    SERVICE_IDLE_CHECK_INTERVAL: timedelta = timedelta(minutes=1)

    # Ensure host data root path ends with a slash
    @validator("HOST_DATA_ROOT_PATH")
    def _validate_host_data_root_path(cls, host_data_root_path: str) -> str:
        if host_data_root_path is None or host_data_root_path.endswith("/"):
            return host_data_root_path
        return f"{host_data_root_path}/"

    # Postgres Connection URI to use for JSON Document Manager
    # If `None`, a dedicated postgres instance will be started as a service (container).
    POSTGRES_CONNECTION_URI: Optional[PostgresDsn] = None

    # S3 Storage Connection Configuration for File Manager
    # If `S3_ENDPOINT` is `None`, a dedicated minio instance will be started as a service (container).
    S3_ENDPOINT: Optional[str] = None
    S3_ACCESS_KEY: Optional[str] = None
    S3_SECRET_KEY: Optional[str] = None
    S3_REGION: Optional[str] = None
    S3_SECURE: Optional[bool] = None
    # TODO: support for managing all data in a single bucket S3_BUCKET: Optional[str] = None

    # Azure Blob Storage Connection Configuration for File Manager
    AZURE_BLOB_CONNECTION_STRING: Optional[str] = None
    AZURE_BLOB_TOKEN: Optional[str] = None

    # Cache Settings
    # VERIFY_ACCESS_CACHE caches all token <-> permission verifications
    VERIFY_ACCESS_CACHE_ENABLED: bool = True  # Enable or disable the cache
    VERIFY_ACCESS_CACHE_SIZE: int = 10000  # number of items in the cache
    VERIFY_ACCESS_CACHE_EXPIRY: int = 300  # Time to live of cache items in seconds
    # API_TOKEN_CACHE caches all db request to get the api token metadata
    API_TOKEN_CACHE_ENABLED: bool = True  # Enable or disable the cache
    API_TOKEN_CACHE_SIZE: int = 10000  # number of items in the cache
    API_TOKEN_CACHE_EXPIRY: int = 300  # Time to live of cache items in seconds
    # RESOURCE_PERMISSIONS caches all permissions granted to resources.
    RESOURCE_PERMISSIONS_CACHE_ENABLED: bool = (
        False  # Enable or disable the cache # TODO: currently not working -> deadlock
    )
    RESOURCE_PERMISSIONS_CACHE_SIZE: int = 10000  # number of items in the cache
    RESOURCE_PERMISSIONS_CACHE_EXPIRY: int = 10  # Time to live of cache items in seconds - This cache should have a very short lifetime

    # Usabel to deactivate setting or changing user passwords
    # The `system-admin` account can still set and change passwords for users,
    # or use the basic auth authentication login page
    # TODO: Maybe only introduce (rename?) a flag indicating whether an external identity provider based on OIDC is used for authentication or whether the default "contaxy password flow" will be used
    PASSWORD_AUTH_ENABLED: bool = True
    USER_REGISTRATION_ENABLED: bool = True

    # External Identity provider configuration
    # ! To test this locally OAUTHLIB_INSECURE_TRANSPORT=1 needs to be set as env variable
    OIDC_AUTH_ENABLED: bool = False
    OIDC_AUTH_URL: Optional[str] = None
    OIDC_TOKEN_URL: Optional[str] = None
    OIDC_CLIENT_ID: Optional[str] = None
    OIDC_CLIENT_SECRET: Optional[str] = None

    # JWT Session Tokens
    JWT_TOKEN_SECRET: str = "please-change-this-secret"
    JWT_TOKEN_EXPIRY_MINUTES: int = 15
    JWT_ALGORITHM: str = "HS256"

    # API Token Length
    API_TOKEN_LENGTH: int = 40

    # BACKEND_CORS_ORIGINS is a JSON-formatted list of origins
    # e.g: '["http://localhost", "http://localhost:4200", "http://localhost:3000", \
    # "http://localhost:8080", "http://local.dockertoolbox.tiangolo.com"]'
    # see this GitHub issue for issue with env list decoding and why validator is needed: https://github.com/samuelcolvin/pydantic/issues/1458
    BACKEND_CORS_ORIGINS: Union[str, List[AnyHttpUrl]] = []

    @validator("BACKEND_CORS_ORIGINS", pre=True, allow_reuse=True)
    def _assemble_cors_origins(
        cls, cors_origins: Union[str, List[AnyHttpUrl]]
    ) -> Union[str, List[AnyHttpUrl]]:
        if isinstance(cors_origins, str):
            return [item.strip() for item in cors_origins.split(",")]  # type: ignore
        return cors_origins

    # TODO: Finalize
    DEBUG: bool = True
    DEBUG_DEACTIVATE_VERIFICATION: bool = False

    def get_redirect_uri(self, omit_host: bool = False) -> str:
        """Get the redirect URI composed of the CONTAXY_HOST and CONTAXY_API_BASE_URL."""

        if not omit_host and not self.CONTAXY_HOST:
            logger.critical(
                "The CONTAXY_HOST configuration is missing and OIDC_AUTH_ENABLED."
            )
            return ""

        if omit_host:
            host = ""
            base_url = "" if not self.CONTAXY_BASE_URL else self.CONTAXY_BASE_URL
        else:
            assert self.CONTAXY_HOST
            host = self.CONTAXY_HOST
            base_url = (
                "" if not self.CONTAXY_BASE_URL else self.CONTAXY_BASE_URL.lstrip("/")
            )

        return os.path.join(host, base_url)

    class Config:
        # Support local .env files
        env_file = ".env"
        env_file_encoding = "utf-8"
        # TODO: Support docker secrets
        # Only newer versions will throw warnings
        # secrets_dir = "/run/secrets"


settings = Settings()



from typing import Any

from fastapi import APIRouter, Depends

from contaxy.api.dependencies import ComponentManager, get_component_manager
from contaxy.schema import AccessLevel
from contaxy.schema.exceptions import VALIDATION_ERROR_RESPONSE, ClientBaseError

router = APIRouter(tags=["seed"], responses={**VALIDATION_ERROR_RESPONSE})


@router.get("/seed/default")
def create_seed_default(
    component_manager: ComponentManager = Depends(get_component_manager),
    # token: str = Depends(get_api_token),
) -> Any:
    # Don't use `verify_token` because that function wants to query the database and right now we don't have a user in the database on startup
    # access_token = component_manager.get_auth_manager()._resolve_token(
    #     token, use_cache=True
    # )
    # if "system#admin" not in access_token.scopes:
    #     raise ClientBaseError(
    #         403,
    #         message="No permission to create seed.",
    #         explanation="Only tokens with `system#admin` scope are permitted.",
    #     )

    seed_manager = component_manager.get_seed_manager()
    user = seed_manager.create_user()
    project = seed_manager.create_project()
    if not project or not project.id:
        raise ClientBaseError(status_code=500, message="Could not create seed data")
    component_manager.get_project_manager().add_project_member(
        project_id=project.id, user_id=user.id, access_level=AccessLevel.ADMIN
    )
    seed_manager.create_file(project_id=project.id)



import json
from typing import Any, List, Optional

from loguru import logger

from contaxy import __version__, config
from contaxy.config import settings
from contaxy.operations import AuthOperations, SystemOperations
from contaxy.operations.components import ComponentOperations
from contaxy.operations.json_db import JsonDocumentOperations
from contaxy.operations.project import ProjectOperations
from contaxy.schema import (
    ClientValueError,
    ResourceAlreadyExistsError,
    ResourceNotFoundError,
)
from contaxy.schema.auth import (
    ADMIN_ROLE,
    USER_ROLE,
    USERS_KIND,
    AccessLevel,
    UserRegistration,
)
from contaxy.schema.extension import GLOBAL_EXTENSION_PROJECT
from contaxy.schema.project import PROJECTS_KIND, ProjectCreation
from contaxy.schema.system import (
    AllowedImageInfo,
    SystemInfo,
    SystemState,
    SystemStatistics,
)
from contaxy.utils import auth_utils


class SystemManager(SystemOperations):
    _ALLOWED_IMAGES_COLLECTION = "allowed-images"

    _SYSTEM_PROPERTIES_COLLECTION = "system-properties"
    _SYSTEM_PROPERTY_IS_INITIALIZED = "is-initialized"

    def __init__(
        self,
        component_manager: ComponentOperations,
    ):
        """Initializes the system manager.

        Args:
            component_manager: Instance of the component manager that grants access to the other managers.
        """
        self._global_state = component_manager.global_state
        self._request_state = component_manager.request_state
        self._component_manager = component_manager

    @property
    def _json_db_manager(self) -> JsonDocumentOperations:
        return self._component_manager.get_json_db_manager()

    @property
    def _auth_manager(self) -> AuthOperations:
        return self._component_manager.get_auth_manager()

    @property
    def _project_manager(self) -> ProjectOperations:
        return self._component_manager.get_project_manager()

    def get_system_info(self) -> SystemInfo:
        if self._is_initialized():
            system_state = SystemState.RUNNING
        else:
            system_state = SystemState.UNINITIALIZED
        return SystemInfo(
            version=__version__,
            namespace=settings.SYSTEM_NAMESPACE,
            system_state=system_state,
        )

    def is_healthy(self) -> bool:
        # TODO: do real healthchecks
        return True

    def get_system_statistics(self) -> SystemStatistics:
        # TODO: Implement system statistics
        return SystemStatistics(
            project_count=0, user_count=0, job_count=0, service_count=0, file_count=0
        )

    def initialize_system(
        self,
        password: Optional[str] = None,
    ) -> None:
        # Only allow initialization once
        if self._is_initialized():
            raise ResourceAlreadyExistsError("The system has already been initialized")
        logger.debug("Started system initialization.")

        # Remove authorized access info
        self._request_state.authorized_access = None

        # Create System Internal Project
        logger.debug("System initialization: Create system project.")
        self._project_manager.create_project(
            ProjectCreation(
                id=config.SYSTEM_INTERNAL_PROJECT,
                display_name=config.SYSTEM_INTERNAL_PROJECT,
                description="This project holds all system internal services and data.",
            ),
            technical_project=True,
        )

        # Create admin role
        logger.debug("System initialization: Creating admin role.")
        self._auth_manager.add_permission(
            ADMIN_ROLE,
            auth_utils.construct_permission("*", AccessLevel.ADMIN),
        )

        # Create user role
        logger.debug("System initialization: Creating user role.")
        # Users needs to know which images are allowed for deployments
        self._auth_manager.add_permission(
            USER_ROLE,
            auth_utils.construct_permission("system/allowed-images", AccessLevel.READ),
        )
        # Users requires read access to the global extension project to access the extension endpoints
        self._auth_manager.add_permission(
            USER_ROLE,
            auth_utils.construct_permission(
                f"{PROJECTS_KIND}/{GLOBAL_EXTENSION_PROJECT}", AccessLevel.READ
            ),
        )

        # Create Admin User
        logger.debug("System initialization: Creating default admin user.")
        # Set initial user password -> SHOULD be changed after the first login
        admin_user = auth_utils.create_and_setup_user(
            user_input=UserRegistration(
                username=config.SYSTEM_ADMIN_USERNAME,
                password=password or config.SYSTEM_ADMIN_INITIAL_PASSWORD,  # type: ignore
            ),
            technical_user=True,
            auth_manager=self._auth_manager,
            project_manager=self._project_manager,
        )
        # Add admin role to admin user
        # TODO: use resource name
        admin_user_resource_name = USERS_KIND + "/" + admin_user.id
        self._auth_manager.add_permission(
            admin_user_resource_name,
            ADMIN_ROLE,
        )

        self._set_system_property(SystemManager._SYSTEM_PROPERTY_IS_INITIALIZED, True)

    def check_allowed_image(self, image_name: str, image_tag: str) -> None:
        # If allowed image list is empty (default), then allow all images
        if len(self.list_allowed_images()) == 0:
            return
        try:
            allowed_image_info = self.get_allowed_image(image_name)
        except ResourceNotFoundError:
            raise ClientValueError(
                f"Image {image_name} is not on the list of allowed images!"
            )

        if not (
            "*" in allowed_image_info.image_tags
            or image_tag in allowed_image_info.image_tags
        ):
            raise ClientValueError(
                f"Image {image_name} is on the list of allowed images but tag {image_tag} is not allowed!"
            )

    def add_allowed_image(self, allowed_image: AllowedImageInfo) -> AllowedImageInfo:
        allowed_image_doc = self._json_db_manager.create_json_document(
            config.SYSTEM_INTERNAL_PROJECT,
            self._ALLOWED_IMAGES_COLLECTION,
            key=allowed_image.image_name,
            json_document=allowed_image.json(),
            upsert=True,
        )
        return AllowedImageInfo.parse_raw(allowed_image_doc.json_value)

    def list_allowed_images(self) -> List[AllowedImageInfo]:
        allowed_image_docs = self._json_db_manager.list_json_documents(
            config.SYSTEM_INTERNAL_PROJECT,
            self._ALLOWED_IMAGES_COLLECTION,
        )
        return [
            AllowedImageInfo.parse_raw(doc.json_value) for doc in allowed_image_docs
        ]

    def get_allowed_image(self, image_name: str) -> AllowedImageInfo:
        allowed_image_doc = self._json_db_manager.get_json_document(
            config.SYSTEM_INTERNAL_PROJECT,
            self._ALLOWED_IMAGES_COLLECTION,
            key=image_name,
        )
        return AllowedImageInfo.parse_raw(allowed_image_doc.json_value)

    def delete_allowed_image(self, image_name: str) -> None:
        self._json_db_manager.delete_json_document(
            config.SYSTEM_INTERNAL_PROJECT,
            self._ALLOWED_IMAGES_COLLECTION,
            key=image_name,
        )

    _is_initialized_cache = False

    def _is_initialized(self) -> bool:
        # After the initialization was successful, is_initialized cannot be set back to false, so caching it is safe
        if not SystemManager._is_initialized_cache:
            # Cache says that system is not initialized. Check if that is correct.
            SystemManager._is_initialized_cache = self._get_system_property(
                self._SYSTEM_PROPERTY_IS_INITIALIZED, default=False
            )
        return SystemManager._is_initialized_cache

    def _get_system_property(self, property_name: str, default: Any = ...) -> Any:
        """Returns the value of the system property.

        Args:
            property_name (str): Name of the system property that should be retrieved
            default (None): Default value to return if property is not set. If not set, ResourceNotFoundError will be raised instead.

        Raises:
            ResourceNotFoundError: If no system property with the given name is found

        Returns:
            The value of the system property
        """
        try:
            system_property_doc = self._json_db_manager.get_json_document(
                config.SYSTEM_INTERNAL_PROJECT,
                self._SYSTEM_PROPERTIES_COLLECTION,
                key=property_name,
            )
        except ResourceNotFoundError as e:
            if default is ...:
                raise e
            else:
                return default
        is_initialized = json.loads(system_property_doc.json_value)
        return is_initialized

    def _set_system_property(self, property_name: str, property_value: Any) -> None:
        """Sets the specified value for the given system property.

        Args:
            property_name (str): The name of the property to set
            property_value (Any): The new value of the property
        """
        self._json_db_manager.create_json_document(
            config.SYSTEM_INTERNAL_PROJECT,
            self._SYSTEM_PROPERTIES_COLLECTION,
            key=property_name,
            json_document=json.dumps(property_value),
            upsert=True,
        )



from typing import Dict, List

import requests
from pydantic.tools import parse_raw_as

from contaxy.clients.shared import handle_errors
from contaxy.operations.system import SystemOperations
from contaxy.schema.system import AllowedImageInfo, SystemInfo, SystemStatistics


class SystemClient(SystemOperations):
    def __init__(self, client: requests.Session):
        self._client = client

    def get_system_info(self, request_kwargs: Dict = {}) -> SystemInfo:
        response = self._client.get("/system/info", **request_kwargs)
        handle_errors(response)
        return parse_raw_as(SystemInfo, response.text)

    def is_healthy(self, request_kwargs: Dict = {}) -> bool:
        response = self._client.get("/system/health", **request_kwargs)
        handle_errors(response)
        return True

    def get_system_statistics(self, request_kwargs: Dict = {}) -> SystemStatistics:
        response = self._client.get("/system/statistics", **request_kwargs)
        handle_errors(response)
        return parse_raw_as(SystemStatistics, response.text)

    def initialize_system(self, request_kwargs: Dict = {}) -> None:
        response = self._client.post("/system/initialize", **request_kwargs)
        handle_errors(response)

    def list_allowed_images(self, request_kwargs: Dict = {}) -> List[AllowedImageInfo]:
        response = self._client.get("/system/allowed-images", **request_kwargs)
        handle_errors(response)
        return parse_raw_as(List[AllowedImageInfo], response.text)

    def add_allowed_image(
        self, allowed_image: AllowedImageInfo, request_kwargs: Dict = {}
    ) -> AllowedImageInfo:
        resource = self._client.post(
            "/system/allowed-images",
            data=allowed_image.json(exclude_unset=True),
            **request_kwargs,
        )
        handle_errors(resource)
        return parse_raw_as(AllowedImageInfo, resource.text)

    def delete_allowed_image(
        self, allowed_image_name: str, request_kwargs: Dict = {}
    ) -> None:
        response = self._client.delete(
            "/system/allowed-images",
            params={"image_name": allowed_image_name},
            **request_kwargs,
        )
        handle_errors(response)

    def check_allowed_image(self, image_name: str, image_tag: str) -> None:
        # TODO: Implement
        pass



from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Mapping, Optional, Union

from fastapi import status
from pydantic import BaseModel, Field

RESOURCE_ID_REGEX = r"^(?!.*--)[a-zA-Z0-9](?:[a-zA-Z0-9\-]{0,61}[a-zA-Z0-9])?$"
QUALIFIED_RESOURCE_ID_REGEX = (
    r"^(?!.*--)[a-z0-9\-]{1,61}[a-z0-9]$"  # TODO: start with [a-z]
)
MAX_DESCRIPTION_LENGTH = 240
MIN_DISPLAY_NAME_LENGTH = 4
MAX_DISPLAY_NAME_LENGTH = 128


class CoreOperations(str, Enum):
    # Project Endpoints
    LIST_PROJECTS = "list_projects"
    CREATE_PROJECT = "create_project"
    GET_PROJECT = "get_project"
    UPDATE_PROJECT = "update_project"
    DELETE_PROJECT = "delete_project"
    GET_PROJECT_TOKEN = "get_project_token"
    SUGGEST_PROJECT_ID = "suggest_project_id"
    LIST_PROJECT_MEMBERS = "list_project_members"
    ADD_PROJECT_MEMBER = "add_project_member"
    REMOVE_PROJECT_MEMBER = "remove_project_member"
    # User Endpoints
    LIST_USERS = "list_users"
    GET_MY_USER = "get_my_user"
    GET_USER = "get_user"
    UPDATE_USER = "update_user"
    DELETE_USER = "delete_user"
    GET_USER_TOKEN = "get_user_token"
    CREATE_USER = "create_user"
    CHANGE_PASSWORD = "change_password"
    # Auth Endpoints
    REFRESH_TOKEN = "refresh_token"
    VERIFY_ACCESS = "verify_access"
    LIST_API_TOKENS = "list_api_tokens"
    CREATE_TOKEN = "create_token"
    LOGOUT_USER_SESSION = "logout_user_session"
    LOGIN_USER_SESSION = "login_user_session"
    LIST_RESOURCE_PERMISSIONS = "list_resource_permissions"
    SET_RESOURCE_PERMISSIONS = "set_resource_permissions"
    DELETE_RESOURCE_PERMISSIONS = "delete_resource_permissions"
    GET_RESOURCES_WITH_PERMISSION = "get_resources_with_permission"
    # OAuth2 Endpoints
    AUTHORIZE_CLIENT = "authorize_client"
    REQUEST_TOKEN = "request_token"
    REVOKE_TOKEN = "revoke_token"
    INTROSPECT_TOKEN = "introspect_token"
    LOGIN_CALLBACK = "login_callback"
    OAUTH_ENABLED = "oauth_enabled"
    GET_USERINFO = "get_userinfo"
    # Extension Endpoints
    INSTALL_EXTENSION = "install_extension"
    LIST_EXTENSIONS = "list_extensions"
    GET_EXTENSION_METADATA = "get_extension_metadata"
    DELETE_EXTENSION = "delete_extension"
    SUGGEST_EXTENSION_CONFIG = "suggest_extension_config"
    SET_EXTENSION_DEFAULTS = "set_extension_defaults"
    GET_EXTENSION_DEFAULTS = "get_extension_defaults"
    # System Endpoints
    GET_SYSTEM_INFO = "get_system_info"
    GET_SYSTEM_STATISTICS = "get_system_statistics"
    LIST_SYSTEM_NODES = "list_system_nodes"
    INITIALIZE_SYSTEM = "initialize_system"
    CLEANUP_SYSTEM = "cleanup_system"
    ADD_ALLOWED_IMAGE = "add_allowed_image"
    LIST_ALLOWED_IMAGES = "list_allowed_images"
    GET_ALLOWED_IMAGE = "get_allowed_image"
    DELETE_ALLOWED_IMAGE = "delete_allowed_image"
    # Secrets Endpoints
    CREATE_SECRET = "create_secret"
    DELETE_SECRET = "delete_secret"
    LIST_SECRETS = "list_secrets"
    # Configuration Endpoints
    LIST_CONFIGURATIONS = "list_configurations"
    CREATE_CONFIGURATION = "create_configuration"
    UPDATE_CONFIGURATION = "update_configuration"
    DELETE_CONFIGURATION = "delete_configuration"
    GET_CONFIGURATION = "get_configuration"
    GET_CONFIGURATION_PARAMETER = "get_configuration_parameter"
    # JSON Document Endpoints
    LIST_JSON_DOCUMENTS = "list_json_documents"
    CREATE_JSON_DOCUMENT = "create_json_document"
    UPDATE_JSON_DOCUMENT = "update_json_document"
    DELETE_JSON_DOCUMENT = "delete_json_document"
    DELETE_JSON_COLLECTION = "delete_json_collection"
    DELETE_JSON_COLLECTIONS = "delete_json_collections"
    GET_JSON_DOCUMENT = "get_json_document"
    # Service Endpoints
    GET_SERVICE_ACCESS_TOKEN = "get_service_access_token"


class ExtensibleOperations(str, Enum):
    # Service Endpoints
    LIST_SERVICES = "list_services"
    GET_SERVICE_METADATA = "get_service_metadata"
    LIST_DEPLOY_SERVICE_ACTIONS = "list_deploy_service_actions"
    DEPLOY_SERVICE = "deploy_service"
    DELETE_SERVICE = "delete_service"
    DELETE_SERVICES = "delete_services"
    UPDATE_SERVICE = "update_service"
    UPDATE_SERVICE_ACCESS = "update_service_access"
    GET_SERVICE_LOGS = "get_service_logs"
    LIST_SERVICE_ACTIONS = "list_service_actions"
    EXECUTE_SERVICE_ACTION = "execute_service_action"
    SUGGEST_SERVICE_CONFIG = "suggest_service_config"
    ACCESS_SERVICE = "access_service"
    # Job Endpoints
    LIST_JOBS = "list_jobs"
    GET_JOB_METADATA = "get_job_metadata"
    LIST_DEPLOY_JOB_ACTIONS = "list_deploy_job_actions"
    LIST_JOB_ACTIONS = "list_job_actions"
    EXECUTE_JOB_ACTION = "execute_job_action"
    SUGGEST_JOB_CONFIG = "suggest_job_config"
    DEPLOY_JOB = "deploy_job"
    DELETE_JOB = "delete_job"
    DELETE_JOBS = "delete_jobs"
    GET_JOB_LOGS = "get_job_logs"
    # File Endpoints
    LIST_FILES = "list_files"
    GET_FILE_METADATA = "get_file_metadata"
    UPLOAD_FILE = "upload_file"
    UPLOAD_FILE_NO_KEY = "upload_file_no_key"
    DOWNLOAD_FILE = "download_file"
    LIST_FILE_ACTIONS = "list_file_actions"
    EXECUTE_FILE_ACTION = "execute_file_action"
    DELETE_FILE = "delete_file"
    DELETE_FILES = "delete_files"
    GET_FILE_ACCESS_TOKEN = "get_file_access_token"
    UPDATE_FILE_METADATA = "update_file_metadata"
    # Auth Endpoints
    OPEN_LOGIN_PAGE = "open_login_page"


# TODO: https://fastapi.tiangolo.com/advanced/additional-responses/#combine-predefined-responses-and-custom-ones
# DEFAULT_RESPONSES = {
#    404: {"description": "Item not found"},
#    307: {"description": "Redirecting to another URL"},
#    403: {"description": "Not enough privileges"},
# }
# TODO: evaluate status codes: 302,303,...
OPEN_URL_REDIRECT: Mapping[Union[int, str], Dict[str, Any]] = {
    status.HTTP_307_TEMPORARY_REDIRECT: {"description": "Redirecting to another URL"}
}


class ResourceMetadata(BaseModel):
    id: str = Field(
        None,
        example="ac9ldprwdi68oihk34jli3kdp",
        description="Resource ID. Identifies a resource in a given context and time, for example, in combination with its type. Used in API operations and/or configuration files.",
    )
    name: Optional[str] = Field(
        None,
        example="resources/ac9ldprwdi68oihk34jli3kdp",
        description="Resource Name. A relative URI-path that uniquely identifies a resource within the system. Assigned by the server and read-only.",
    )
    # kind: Optional[str] = Field(
    #    None,
    #    example="resources",
    #    description="Resource Type (Schema). Identifies what kind of resource this is. Assigned by the server and read-only.",
    # )
    created_at: Optional[datetime] = Field(
        None,
        description="Timestamp of the resource creation. Assigned by the server and read-only.",
    )
    created_by: Optional[str] = Field(
        None,
        example="resources/ac9ldprwdi68oihk34jli3kdp",
        description="Resource name of the entity responsible for the creation of this resource. Assigned by the server and read-only.",
    )
    updated_at: Optional[datetime] = Field(
        None,
        description="Timestamp of the last resource modification. Is updated when create/patch/delete operation is performed. Assigned by the server and read-only.",
    )
    updated_by: Optional[str] = Field(
        None,
        example="resources/ac9ldprwdi68oihk34jli3kdp",
        description="Resource name of the entity responsible for the last modification of this resource. Assigned by the server and read-only.",
    )
    # TODO: v2
    # global_id: Optional[str] = Field(
    #    None,
    #    description="Globally unique ID of this resource.",
    # )
    # TODO: v2
    # deleted_at: Optional[datetime] = Field(
    #    None,
    #    description="Timestamp when the resource should be deleted.",
    # )


class ResourceInput(BaseModel):
    display_name: Optional[str] = Field(
        None,
        max_length=MAX_DISPLAY_NAME_LENGTH,
        description="A user-defined human-readable name of the resource. The name can be up to 128 characters long and can consist of any UTF-8 character.",
    )
    description: str = Field(
        "",
        max_length=MAX_DESCRIPTION_LENGTH,
        description="A user-defined short description about the resource. Can consist of any UTF-8 character.",
    )
    icon: Optional[str] = Field(
        None,
        description="Identifier or image URL used for displaying this resource.",
    )
    metadata: Dict[str, str] = Field(
        {},
        example={"additional-metadata": "value"},
        description="A collection of arbitrary key-value pairs associated with this resource that does not need predefined structure. Enable third-party integrations to decorate objects with additional metadata for their own use.",
    )
    disabled: bool = Field(
        False,
        description="Allows to disable a resource without requiring deletion. A disabled resource is not shown and not accessible.",
    )
    # TODO: v2
    # tags: Optional[List[str]] = Field(
    #    None,
    #    description="List of tags associated with this resource.",
    # )


class Resource(ResourceInput, ResourceMetadata):
    pass


class ActionInstruction(str, Enum):
    NEW_TAB = "new-tab"


class ResourceAction(BaseModel):
    action_id: Optional[str] = Field(
        ...,
        example="access-8080",
        description="ID used to identify this action.",
        regex=RESOURCE_ID_REGEX,
    )
    display_name: Optional[str] = Field(
        None,
        min_length=MIN_DISPLAY_NAME_LENGTH,
        max_length=MAX_DISPLAY_NAME_LENGTH,
        example="Access Port 8080",
        description="Display name of this action.",
    )
    icon: Optional[str] = Field(
        None,
        max_length=1000,
        example="open_in_new",
        description="Material Design Icon name or image URL used for displaying this action.",
    )
    extension_id: Optional[str] = Field(
        None,
        description="The extension ID associated with this action.",
    )
    extension_name: Optional[str] = Field(
        None,
        description="The extension name associated with this action.",
    )
    instructions: Optional[List[Any]] = Field(
        None,
        example=["new-tab"],
        description="A list of instructions for the frontend application.",
    )


class ResourceActionExecution(BaseModel):
    parameters: Dict[str, str] = Field(
        {},
        description="Parameters that are passed to the resource action.",
        example={"action-parameter": "parameter-value"},
    )


# TODO: use?
# class BaseEntity(BaseModel):
#     class Config:
#        anystr_strip_whitespace = True
#        underscore_attrs_are_private = True
# TODO: should we set a max length for string to prevent to big requests?
#        max_anystr_length = 20000



"""Collection of utilities for FastAPI apps."""

import asyncio
import inspect
from datetime import timedelta
from typing import Any, Callable, Type

from fastapi import FastAPI, Form
from loguru import logger
from pydantic import BaseModel
from starlette.concurrency import run_in_threadpool


def schedule_call(func: Callable, interval: timedelta) -> None:
    """Schedule a function to be called in regular intervals.

    Args:
        func (Callable): The function to be called regularly
        interval (timedelta): The amount of time to wait between function calls
    """

    async def loop() -> None:
        while True:
            await asyncio.sleep(interval.seconds)
            try:
                await run_in_threadpool(func)
            except Exception:
                logger.exception(f"Error while executing scheduled function {func}!")

    asyncio.create_task(loop())


def as_form(cls: Type[BaseModel]) -> Any:
    """Adds an as_form class method to decorated models.

    The as_form class method can be used with FastAPI endpoints
    """
    new_params = [
        inspect.Parameter(
            field.alias,
            inspect.Parameter.POSITIONAL_ONLY,
            default=(Form(field.default) if not field.required else Form(...)),
        )
        for field in cls.__fields__.values()
    ]

    async def _as_form(**data):  # type: ignore
        return cls(**data)

    sig = inspect.signature(_as_form)
    sig = sig.replace(parameters=new_params)
    _as_form.__signature__ = sig  # type: ignore
    setattr(cls, "as_form", _as_form)
    return cls


def patch_fastapi(app: FastAPI) -> None:
    """Patch function to allow relative url resolution.

    This patch is required to make fastapi fully functional with a relative url path.
    This code snippet can be copy-pasted to any Fastapi application.
    """
    from fastapi.openapi.docs import get_redoc_html, get_swagger_ui_html
    from starlette.requests import Request
    from starlette.responses import HTMLResponse

    async def redoc_ui_html(req: Request) -> HTMLResponse:
        assert app.openapi_url is not None
        redoc_ui = get_redoc_html(
            openapi_url="./" + app.openapi_url.lstrip("/"),
            title=app.title + " - Redoc UI",
        )

        return HTMLResponse(redoc_ui.body.decode("utf-8"))

    async def swagger_ui_html(req: Request) -> HTMLResponse:
        assert app.openapi_url is not None
        swagger_ui = get_swagger_ui_html(
            openapi_url="./" + app.openapi_url.lstrip("/"),
            title=app.title + " - Swagger UI",
            oauth2_redirect_url=app.swagger_ui_oauth2_redirect_url,
        )

        # insert request interceptor to have all request run on relativ path
        request_interceptor = (
            "requestInterceptor: (e)  => {"
            "\n\t\t\tvar url = window.location.origin + window.location.pathname"
            '\n\t\t\turl = url.substring( 0, url.lastIndexOf( "/" ) + 1);'
            "\n\t\t\turl = e.url.replace(/http(s)?:\/\/[^/]*\//i, url);"  # noqa: W605
            "\n\t\t\te.contextUrl = url"
            "\n\t\t\te.url = url"
            "\n\t\t\treturn e;}"
        )

        return HTMLResponse(
            swagger_ui.body.decode("utf-8").replace(
                "SwaggerUIBundle({",
                "SwaggerUIBundle({\n\t\t" + request_interceptor + ",",
            )
        )

    # remove old docs route and add our patched route
    routes_new = []
    for app_route in app.routes:
        if app_route.path == "/docs":  # type: ignore
            continue

        if app_route.path == "/redoc":  # type: ignore
            continue

        routes_new.append(app_route)

    app.router.routes = routes_new

    assert app.docs_url is not None
    app.add_route(app.docs_url, swagger_ui_html, include_in_schema=False)
    assert app.redoc_url is not None
    app.add_route(app.redoc_url, redoc_ui_html, include_in_schema=False)


def add_timing_info(app: FastAPI) -> None:
    import logging

    from fastapi_utils.timing import add_timing_middleware

    logging.basicConfig(level=logging.INFO)
    logger = logging.getLogger(__name__)
    add_timing_middleware(app, record=logger.info, prefix="app", exclude="untimed")

    # @app.middleware("http")
    # async def add_process_time_header(request: Request, call_next):
    #    start_time = time.time()
    #    response = await call_next(request)
    #    process_time = time.time() - start_time
    #    response.headers["X-Process-Time"] = str(process_time)
    #    return response
    # fastapi-profiler
    # from fastapi_profiler.profiler_middleware import PyInstrumentProfilerMiddleware
    # app.add_middleware(PyInstrumentProfilerMiddleware, profiler_output_type="html")
    # import fastapi
    # from asgi_server_timing import ServerTimingMiddleware
    # app.add_middleware(
    #    ServerTimingMiddleware,
    #    calls_to_track={
    #        "1deps": (fastapi.routing.solve_dependencies,),
    #        "2main": (fastapi.routing.run_endpoint_function,),
    # "3valid": (pydantic.fields.ModelField.validate,),
    #        "4encode": (fastapi.encoders.jsonable_encoder,),
    #        "5render": (
    #            fastapi.responses.JSONResponse.render,
    #            fastapi.responses.ORJSONResponse.render,
    #        ),
    #    },
    # )


# @app.middleware("http")
# async def add_process_time_header(request: Request, call_next) -> Response:  # type: ignore
#    profiler = Profiler()
#    profiler.start()
#    start_time = time.time()
#
#    response = await call_next(request)
#    process_time = time.time() - start_time
#    response.headers["X-Process-Time"] = str(process_time)
#    profiler.stop()
#
#    with open(
#        "./prof/" + id_utils.generate_readable_id(str(request.url)) + ".html", "w"
#    ) as f:
#        f.write(profiler.output_html())
#    return response
